// file      : agent.cpp
// author    : Boris Kolpackov <boris@dre.vanderbilt.edu>
// cvs-id    : $Id$

#include <iostream>

#include <ace/OS.h>

#include <tao/corba.h>

#include <orbsvcs/CosNotificationC.h>
#include <orbsvcs/CosNotifyChannelAdminC.h>
#include <orbsvcs/CosNotifyCommC.h>
#include <orbsvcs/CosNotifyCommS.h>

#include "Gate.h"

using std::cerr;
using std::endl;

using namespace CORBA;
using namespace CosNotifyComm;
using namespace CosNotification;
using namespace CosNotifyChannelAdmin;

class Agent : public POA_CosNotifyComm::StructuredPushConsumer,
              public PortableServer::RefCountServantBase
{
public:
  Agent (char const* name, EventChannel_ptr ch)
      : name_ (name), counter_ (0)
  {
    // Obtain proxy consumer.
    //
    ProxyConsumer_var pc (
      ch->default_supplier_admin ()->obtain_notification_push_consumer (
        STRUCTURED_EVENT, consumer_id_));

    consumer_ = StructuredProxyPushConsumer::_narrow (pc.in ());

    consumer_->connect_structured_push_supplier (
      StructuredPushSupplier::_nil ());

    // Register as consumer.
    //
    StructuredPushConsumer_var ref (_this ()); // Activate on default POA.

    ProxySupplier_var ps (
      ch->default_consumer_admin ()->obtain_notification_push_supplier (
        STRUCTURED_EVENT, supplier_id_));

    supplier_ = StructuredProxyPushSupplier::_narrow (ps.in ());

    supplier_->connect_structured_push_consumer (ref.in ());

    //
    //
    if (ACE_OS::thr_create (&tracker_thunk,
                            this,
                            THR_JOINABLE,
                            &thread_) != 0) ::abort ();
  }

private:
  static ACE_THR_FUNC_RETURN
  tracker_thunk (void* arg)
  {
    Agent* a (reinterpret_cast<Agent*> (arg));
    a->tracker ();
    return 0;
  }

  void
  tracker ()
  {
    StructuredEvent e;

    // Header.
    //
    e.header.fixed_header.event_type.domain_name = string_dup("SpaceCraft");
    e.header.fixed_header.event_type.type_name = string_dup("AgentDiscovery");
    e.header.fixed_header.event_name = name_;

    while (true)
    {
      e.remainder_of_body <<= counter_++;

      consumer_->push_structured_event (e);

      ACE_OS::sleep (ACE_Time_Value (3, 0));
    }
  }

private:
  // NotifyPublish interface.
  //
  virtual void
  offer_change (EventTypeSeq const&, EventTypeSeq const&) throw ()
  {
    // We don't care.
  }

  // StructuredPushSupplier interface.
  //
  virtual void
  push_structured_event (StructuredEvent const& e) throw ()
  {
    CORBA::ULong counter;

    e.remainder_of_body >>= counter;

    cerr << e.header.fixed_header.event_type.domain_name << "::"
         << e.header.fixed_header.event_type.type_name << " "
         << e.header.fixed_header.event_name << " : "
         << counter << endl;
  }


  virtual void
  disconnect_structured_push_consumer () throw ()
  {
    // We don't care.
  }

private:
  String_var name_;
  ULong counter_;

  ACE_thread_t thread_;

  ProxyID consumer_id_;
  StructuredProxyPushConsumer_var consumer_;

  ProxyID supplier_id_;
  StructuredProxyPushSupplier_var supplier_;

};

int
main (int argc, char* argv[])
{
  ORB_var orb (ORB_init (argc, argv));

  if (argc != 2)
  {
    cerr << "usage: " << argv[0] << " <name>" << endl;
    return 1;
  }


  // Activate root POA.
  //
  CORBA::Object_var obj (orb->resolve_initial_references ("RootPOA"));
  PortableServer::POA_var root_poa (PortableServer::POA::_narrow (obj.in ()));

  PortableServer::POAManager_var poa_manager (root_poa->the_POAManager ());

  poa_manager->activate ();


  // Get a channel factory.
  //
  obj = (orb->resolve_initial_references ("Notify"));
  EventChannelFactory_var factory (EventChannelFactory::_narrow (obj.in ()));

  if (is_nil (factory.in ()))
  {
    cerr << "unable to resolve Notification Service reference" << endl;
    return 1;
  }


  // Create a channel.
  //
  QoSProperties qosp;
  AdminProperties ap;
  ChannelID id;

  EventChannel_var channel (factory->create_channel (qosp, ap, id));

  ACE_INET_Addr addr ("224.1.0.1:10000");

  Gate gate (addr, channel);

  Agent smith (argv[1], channel);
  //Agent johnson ("johnson", channel);

  orb->run ();
}
